// name: StdResH.h
// author: J. Michael Word
// date written: 2/8/96 with many later modifications
// purpose: describes the connectivies, angles
//          and distances required to place
//          hydrogens on each standard residue

// **************************************************************
// NOTICE: This is free software and the source code is freely
// available. You are free to redistribute or modify under the
// conditions that (1) this notice is not removed or modified
// in any way and (2) any modified versions of the program are
// also available for free.
//               ** Absolutely no Warranty **
// Copyright (C) 1999 J. Michael Word
// **************************************************************

#ifndef STDRESH_H
#define STDRESH_H 1

#include "AtomConn.h"
#include "Seq.h"
#include "Stringclass.h"
#include "Dict.h"
#include "ElementInfo.h"

// types of hydrogens-
// 1: HXR3 - requires just 4 atom centers
// 2: H2XR2- three atoms and an angle
// 3: H3XR - three atoms, an angle and dihedral
// 4: HXR2 - three atoms and a fudge factor (planar interpolating vectors)
// 5: HXR2 - three atoms and a fraction (planar based on angles)
// 6: HXY  - (linear) just two atoms
// 7: HX   - (can't create, can only adjust length) only one atom

class HydrogenPlanTable;
class StdResXtraInfo;

class StdResH {
public:
   StdResH(const char* rn, const char* ex) : _name(rn), _exclude(ex) {}
   ~StdResH() {}

   void addPlan(int type, const char* elem, const char *hname,
                const char *c1, const char *c2, const char *c3, const char *c4,
                double dist, double ang1, double ang2, int flags);

   Stringclass name() const { return _name; }
   Stringclass exclude() const { return _exclude; }
   Seq<atomPlacementPlan> plans() const { return _plans; }

   static const HydrogenPlanTable& HydPlanTbl();
   static const StdResXtraInfo&    ResXtraInfo();

private:
   static const HydrogenPlanTable *TheHydPlanTbl;    // class shared resource
   static const StdResXtraInfo    *TheResXtraInfo;   // class shared resource

   Stringclass                 _name;    // residue name
   Stringclass                 _exclude; // list of incompatible residues (for n-term, etc.)
   Seq<atomPlacementPlan> _plans;
};

// -----------------------------------------------------------------------------
// implementation class for a table of Hydrogen building records by Residue type
// -----------------------------------------------------------------------------
class HydrogenPlanTable {
public:
   HydrogenPlanTable();
   ~HydrogenPlanTable() {}

   StdResH* get(const Stringclass& resName) const { return _restbl.get(resName); }

private:
   HydrogenPlanTable(const HydrogenPlanTable&);           //can't copy
   HydrogenPlanTable& operator=(const HydrogenPlanTable&);//can't assign

   Dict<Stringclass, StdResH> _restbl;
};

// -----------------------------------------------------------------------------
// supplemental information about the standard residues (currently limited)
// -----------------------------------------------------------------------------
class StdResXtraInfo {
public:
   StdResXtraInfo();
   ~StdResXtraInfo() {}

   bool atomHasAttrib(const Stringclass& resName, const Stringclass& atomName, int attr) const {
      int *p = _atomAttributes.get(makeKey(resName, atomName));
      return (p == NULL) ? FALSE : ((*p) & attr);
   }

   bool match(const Stringclass& s1, const Stringclass& s2) const {
      int *p = _atomAttributes.get(makeKey(s1, s2));
      return (p == NULL) ? FALSE : ((*p) != 0);
   }

private:
   StdResXtraInfo(const StdResXtraInfo&);           //can't copy
   StdResXtraInfo& operator=(const StdResXtraInfo&);//can't assign

   Stringclass makeKey(const Stringclass& resName, const Stringclass& atomName) const {
      return resName + ":" + atomName;
   }

   Dict<Stringclass, int> _atomAttributes;
};
#endif
